# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from types import GenericAlias
from typing import Any, Callable, Union, get_origin

from json_any.catalog.type import enum_t
from json_any.constant import MODULE_TYPE_SEPARATOR
from json_any.extension.module import ElementInModule

builders_h = dict[str, Callable[[Any], Any]]
description_h = tuple[str, Any]
descriptors_h = dict[str, Callable[[Any], Any]]


class unfound_t:
    pass


def QualifiedType(instance_or_type: Union[Any, type], /) -> str:
    """"""
    # TODO: Explain why an enum_t instance must not be processed by TypeOfInstance.
    if (type(instance_or_type) is type) or issubclass(instance_or_type, enum_t):
        type_ = instance_or_type
    else:
        type_ = TypeOfInstance(instance_or_type)

    return f"{type_.__module__}{MODULE_TYPE_SEPARATOR}{type_.__name__}"


def TypeFromJsonType(
    json_type: str, /, *, prefix: str = "", should_continue_on_error: bool = False
) -> type:
    """"""
    module, type_ = json_type[prefix.__len__() :].split(
        MODULE_TYPE_SEPARATOR, maxsplit=1
    )

    output, error = ElementInModule(type_, module)
    if output is None:
        if should_continue_on_error:
            output = unfound_t
        else:
            raise TypeError(error)

    return output


def TypeOfInstance(instance: Any, /) -> type:
    """"""
    if isinstance(instance, GenericAlias):
        return get_origin(type(instance))

    return type(instance)


def IsNamedTuple(instance: Any, /) -> bool:
    """"""
    instance_type = TypeOfInstance(instance)
    if hasattr(instance_type, "_make"):
        try:
            as_tuple = tuple(instance)
        except TypeError:
            return False

        return instance_type._make(as_tuple) == instance

    return False


def IsFullyDataclassBased(instance_type: type, /) -> bool:
    """"""
    if dtcl.is_dataclass(instance_type):
        inheritance = instance_type.__mro__[1:]  # Exclude oneself
        if inheritance[0] is object:
            return True

        # Exclude object
        return all(IsFullyDataclassBased(_typ) for _typ in inheritance[:-1])

    return False



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from io import BytesIO as bytes_io_t
from io import StringIO as string_io_t
from pathlib import Path as path_t
from typing import Any, Callable, Sequence, NoReturn, Hashable

from json_any.extension.type import builders_h, descriptors_h
from json_any.json_to_object import ObjectFromJsonString
from json_any.object_to_json import JsonStringOf
from json_any.task.compression import (
    STANDARD_COMPRESSOR_MODULES,
    CompressedVersion,
    DecompressedVersion,
)

STANDARD_COMPRESSOR_EXTENSIONS = {
    None: ".json",
    "bz2": ".json.bz2",
    "gzip": ".json.gz",
    "lzma": ".json.lzma",
    "zlib": ".json.zlib",
}
DECOMPRESSOR_AUTO = "AUTO"


def StoreAsJSON(
    instance: Any,
    path: str | path_t | bytes_io_t | string_io_t,
    /,
    *args,
    descriptors: descriptors_h = None,
    compressor: str
    | Callable[[bytes, ...], bytes]
    | None = STANDARD_COMPRESSOR_MODULES[0],
    should_continue_on_error: bool = False,
    should_add_standard_extension: bool = True,
    should_overwrite_path: bool = False,
    **kwargs,
) -> path_t | Sequence[str] | None:
    """"""
    if isinstance(path, str):
        path = path_t(path)

    if (
        isinstance(path, path_t)
        and isinstance(compressor, Hashable)
        and (compressor in STANDARD_COMPRESSOR_EXTENSIONS)
    ):
        extensions = path.suffixes
        if extensions.__len__() > 0:
            _CheckCompressorAndStdExtensionMatching(
                compressor, "compressor", extensions
            )
        elif should_add_standard_extension:
            path = path.with_suffix(STANDARD_COMPRESSOR_EXTENSIONS[compressor])

    if (
        isinstance(path, path_t)
        and path.exists()
        and not (path.is_file() and should_overwrite_path)
    ):
        raise ValueError(
            f"{path}: Path exists and is not a file or should not be overwritten."
        )
    if (isinstance(path, bytes_io_t) and (compressor is None)) or (
        isinstance(path, string_io_t) and (compressor is not None)
    ):
        raise ValueError(
            f"T.{type(path).__name__}, C.{compressor}: Path-like type T and "
            f"compression C mismatch. Expected={bytes_io_t} with compression, "
            f"or {string_io_t} without compression."
        )

    jsoned, history = JsonStringOf(instance, descriptors=descriptors)
    if history is None:
        if compressor is None:
            content = jsoned
            mode = "w"
        else:
            content = CompressedVersion(jsoned, *args, compressor=compressor, **kwargs)
            mode = "wb"
        if isinstance(path, path_t):
            with open(path, mode=mode) as json_accessor:
                json_accessor.write(content)
            return path
        else:
            path.write(content)
    elif should_continue_on_error:
        return history
    else:
        raise RuntimeError("\n".join(history))


def LoadFromJSON(
    path: str | path_t | bytes_io_t | string_io_t,
    /,
    *args,
    builders: builders_h = None,
    decompressor: str | Callable[[bytes, ...], bytes] | None = DECOMPRESSOR_AUTO,
    should_continue_on_error: bool = False,
    **kwargs,
) -> Any:
    """"""
    if isinstance(path, str):
        path = path_t(path)

    if isinstance(path, path_t):
        extensions = path.suffixes
        if decompressor == DECOMPRESSOR_AUTO:
            if extensions.__len__() > 0:
                decompressor = _DecompressorFromExtension(extensions)
            else:
                raise ValueError(
                    f"{path.name}: Decompressor selection cannot be left to loader "
                    f"when loading from a path without extension."
                )
        elif (
            (extensions.__len__() > 0)
            and isinstance(decompressor, Hashable)
            and (decompressor in STANDARD_COMPRESSOR_EXTENSIONS)
        ):
            _CheckCompressorAndStdExtensionMatching(
                decompressor, "decompressor", extensions
            )

        if decompressor is None:
            mode = "r"
        else:
            mode = "rb"
        with open(path, mode=mode) as json_accessor:
            content = json_accessor.read()
    elif decompressor == DECOMPRESSOR_AUTO:
        raise ValueError(
            f"{decompressor}: Decompressor selection cannot be left to loader "
            f"when not loading from path."
        )
    else:
        if (isinstance(path, bytes_io_t) and (decompressor is None)) or (
            isinstance(path, string_io_t) and (decompressor is not None)
        ):
            raise ValueError(
                f"T.{type(path).__name__}, D.{decompressor}: Path-like type T and "
                f"decompression D mismatch. Expected={bytes_io_t} with decompression, "
                f"or {string_io_t} without decompression."
            )
        content = path.read()

    if decompressor is None:
        jsoned = content
    else:
        jsoned = DecompressedVersion(
            content,
            *args,
            decompressor=decompressor,
            **kwargs,
        )

    return ObjectFromJsonString(
        jsoned,
        builders=builders,
        should_continue_on_error=should_continue_on_error,
    )


def _CheckCompressorAndStdExtensionMatching(
    compressor: Hashable, actual: str, extensions: Sequence[str], /
) -> None | NoReturn:
    """
    Actually, compressor or decompressor, with "actual" allowing to make the difference.
    """
    full_extension = "".join(extensions)
    found = False
    for (
        std_compressor,
        std_full_extension,
    ) in STANDARD_COMPRESSOR_EXTENSIONS.items():
        if full_extension.lower() == std_full_extension:
            found = compressor == std_compressor
            break

    if not found:
        initial = actual[0].upper()
        expected = " or ".join(
            f"{initial}.{_key}+E.{_vle}"
            for _key, _vle in STANDARD_COMPRESSOR_EXTENSIONS.items()
        )
        raise ValueError(
            f"{initial}.{compressor}, E.{full_extension}: {actual.capitalize()} "
            f"{initial} and extension E do not match. Expected={expected}."
        )


def _DecompressorFromExtension(extensions: Sequence[str], /) -> str | None | NoReturn:
    """"""
    full_extension = "".join(extensions)
    for (
        std_compressor,
        std_full_extension,
    ) in STANDARD_COMPRESSOR_EXTENSIONS.items():
        if full_extension.lower() == std_full_extension:
            return std_compressor

    expected = " or ".join(STANDARD_COMPRESSOR_EXTENSIONS.values())
    raise ValueError(
        f"{full_extension}: Not a valid extension for automatic decompressor "
        f"selection. Expected={expected}."
    )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from array import array as py_array_t
from datetime import date as date_t
from datetime import datetime as date_time_t
from datetime import time as time_t
from datetime import timedelta as time_delta_t
from datetime import timezone as time_zone_t
from enum import Enum as enum_t
from io import BytesIO as io_bytes_t
from pathlib import PurePath as path_t
from typing import NamedTuple as named_tuple_t
from typing import Any, Sequence
from uuid import UUID as uuid_t

from json_any.catalog.module import grph, nmpy, pnds, pypl, sprs
from json_any.constant import MODULE_TYPE_SEPARATOR

BUILTIN_BYTES_CONTAINERS = (bytes, bytearray)
BUILTIN_CONTAINERS = (frozenset, list, set, tuple)
JSON_TYPE: dict[type, str] = {}  # Matches non-builtin types with module-name_type-name.
JSON_TYPE_NUMPY_SCALAR = f"{nmpy.__name__}{MODULE_TYPE_SEPARATOR}SCALAR"
JSON_TYPE_PREFIX_PATHLIB = f"{path_t.__module__}{MODULE_TYPE_SEPARATOR}"

BUILTIN_BYTES_CONTAINERS_NAMES = tuple(
    _tpe.__name__ for _tpe in BUILTIN_BYTES_CONTAINERS
)
BUILTIN_CONTAINERS_NAMES = tuple(_tpe.__name__ for _tpe in BUILTIN_CONTAINERS)


def ContainerWithName(name: str, /) -> type:
    """"""
    if name in BUILTIN_BYTES_CONTAINERS_NAMES:
        return BUILTIN_BYTES_CONTAINERS[BUILTIN_BYTES_CONTAINERS_NAMES.index(name)]
    else:
        return BUILTIN_CONTAINERS[BUILTIN_CONTAINERS_NAMES.index(name)]


def TypeIsInModule(json_type: str, module: Any, /) -> bool:
    """"""
    module_name = module.__name__
    return json_type.startswith(module_name) and (
        json_type[module_name.__len__()] in (".", MODULE_TYPE_SEPARATOR)
    )


def TypeNameOf(json_type: str, /) -> str:
    """"""
    return json_type[(json_type.rindex(MODULE_TYPE_SEPARATOR) + 1) :]


def _AddJsonTypes(types: Sequence[type], /) -> None:
    """"""
    global JSON_TYPE

    for type_ in types:
        JSON_TYPE[type_] = f"{type_.__module__}{MODULE_TYPE_SEPARATOR}{type_.__name__}"


_AddJsonTypes(
    (
        date_t,
        date_time_t,
        enum_t,
        io_bytes_t,
        named_tuple_t,
        py_array_t,
        time_delta_t,
        time_t,
        time_zone_t,
        uuid_t,
    )
)


# Note: When there is a single class of interest in a module, the purpose of a *_CLASSES
# tuple is to have an homogeneous dealing of all modules.

if pypl is None:
    MATPLOTLIB_CLASSES = ()
else:
    figure_t = pypl.Figure
    MATPLOTLIB_CLASSES = (figure_t,)
    # Unfortunately, figure_t.__nodule__ is reported as "matplotlib.figure" instead of
    # the expected "matplotlib.pyplot".
    JSON_TYPE[figure_t] = f"{pypl.__name__}{MODULE_TYPE_SEPARATOR}{figure_t.__name__}"

if grph is None:
    NETWORKX_CLASSES = ()
else:
    NETWORKX_CLASSES = (grph.Graph, grph.DiGraph, grph.MultiGraph, grph.MultiDiGraph)
    _AddJsonTypes(NETWORKX_CLASSES)

if nmpy is None:
    np_array_t = None
    NUMPY_ARRAY_CLASSES = ()
else:
    np_array_t = nmpy.ndarray
    NUMPY_ARRAY_CLASSES = (np_array_t,)
    _AddJsonTypes(NUMPY_ARRAY_CLASSES)

if pnds is None:
    PANDAS_CLASSES = ()
else:
    PANDAS_CLASSES = (pnds.Series, pnds.DataFrame)
    _AddJsonTypes(PANDAS_CLASSES)

if sprs is None:
    SCIPY_ARRAY_CLASSES = ()
else:
    SCIPY_ARRAY_CLASSES = (
        sprs.bsr_array,
        sprs.coo_array,
        sprs.csc_array,
        sprs.csr_array,
        sprs.dia_array,
        sprs.dok_array,
        sprs.lil_array,
    )
    _AddJsonTypes(SCIPY_ARRAY_CLASSES)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import json
from typing import Any, Callable

from json_any.catalog.module import blsc, nmpy, pcst
from json_any.catalog.type import io_bytes_t
from json_any.catalog.type import np_array_t as array_t

BLOSC = "blosc"
BLOSC_BYTES = "blosc.bytes"
NUMPY = "numpy"
NUMPY_COMPRESSED = "numpy.compressed"
PCA_B_STREAM = "pca_b_stream"

_NUMPY_NDARRAY_FROM_VERSIONS = {}
_NUMPY_NDARRAY_TO_VERSIONS = {}

if nmpy is None:

    def _RaiseException() -> None:
        """"""
        raise NotImplementedError(f"Numpy module not installed or importable.")

    def IsNumpyScalar(*_, **__) -> Any:
        _RaiseException()

    def AsScalarRepresentation(*_, **__) -> Any:
        _RaiseException()

    def AsMostConciseRepresentation(*_, **__) -> Any:
        _RaiseException()

    def AsNumpyArray(*_, **__) -> Any:
        _RaiseException()

    def AddNumpyNDArrayRepresentation(*_, **__) -> None:
        _RaiseException()

    def RemoveNumpyNDArrayRepresentation(*_, **__) -> None:
        _RaiseException()

else:
    _SCALAR_META_TYPES = (nmpy.bool_, nmpy.number, nmpy.character)

    def IsNumpyScalar(type_: type, /) -> bool:
        """"""
        return nmpy.issubclass_(type_, _SCALAR_META_TYPES)

    def AsScalarRepresentation(instance: Any, type_: type, /) -> Any:
        """"""
        if nmpy.issubclass_(type_, nmpy.bool_):
            output = bool(instance)
        elif nmpy.issubclass_(type_, nmpy.integer):
            output = int(instance)
        elif nmpy.issubclass_(type_, nmpy.floating):
            output = float(instance)
        elif nmpy.issubclass_(type_, complex):
            # Use list (instead of tuple) since tuples seem to be unJSONed as list, so...
            output = [float(nmpy.real(instance)), float(nmpy.imag(instance))]
        else:
            output = str(instance)

        return output

    def AsMostConciseRepresentation(array: array_t, /) -> tuple[str, str]:
        """"""
        version = (array.tolist(), array.dtype.char)
        try:
            min_length = json.dumps(version).__len__()
            output = (NUMPY, version)
        except TypeError:
            min_length = None
            output = ("None", None)

        fake_file = io_bytes_t()
        nmpy.savez_compressed(fake_file, array=array)
        version = fake_file.getvalue().decode(encoding="iso-8859-1")
        fake_file.close()
        length = version.__len__()
        if (min_length is None) or (length < min_length):
            output, min_length = (NUMPY_COMPRESSED, version), length

        for ToVersion in _NUMPY_NDARRAY_TO_VERSIONS.values():
            version = ToVersion(array)
            if version is None:
                continue
            length = version[1].__len__()
            if length < min_length:
                output, min_length = version, length

        return output

    def AsNumpyArray(how: str, what: str) -> array_t:
        """"""
        if how == NUMPY:
            data, dtype = what
            return nmpy.array(data, dtype=dtype)
        elif how == NUMPY_COMPRESSED:
            fake_file = io_bytes_t(what.encode(encoding="iso-8859-1"))
            output = nmpy.load(fake_file)["array"]
            fake_file.close()
            return output

        return _NUMPY_NDARRAY_FROM_VERSIONS[how](what)

    def AddNumpyNDArrayRepresentation(
        name: str,
        /,
        *,
        ToVersion: Callable[[array_t], tuple[int, str, str]] = None,
        FromVersion: Callable[[str], array_t] = None,
    ) -> None:
        """"""
        global _NUMPY_NDARRAY_TO_VERSIONS, _NUMPY_NDARRAY_FROM_VERSIONS

        if name in (NUMPY, NUMPY_COMPRESSED):
            raise ValueError(
                f"{NUMPY}, {NUMPY_COMPRESSED}: Reserved representation names"
            )

        if name == BLOSC:
            if blsc is None:
                raise ModuleNotFoundError('Module "blosc" not installed or unfoundable')
            _NUMPY_NDARRAY_TO_VERSIONS[BLOSC] = _BloscVersion
            _NUMPY_NDARRAY_FROM_VERSIONS[BLOSC] = _FromBloscVersion
            _NUMPY_NDARRAY_FROM_VERSIONS[BLOSC_BYTES] = _FromBloscBytesVersion
        elif name == PCA_B_STREAM:
            if pcst is None:
                raise ModuleNotFoundError(
                    'Module "pca_b_stream" not installed or unfoundable'
                )
            _NUMPY_NDARRAY_TO_VERSIONS[PCA_B_STREAM] = _PCABStreamVersion
            _NUMPY_NDARRAY_FROM_VERSIONS[PCA_B_STREAM] = _FromPCABStreamVersion
        else:
            if (ToVersion is None) or (FromVersion is None):
                raise ValueError(
                    f'{name}: Invalid keyword-only arguments "ToVersion" and/or "FromVersion". '
                    f"Actual={ToVersion}/{FromVersion}. Expected=Both non-None."
                )
            _NUMPY_NDARRAY_TO_VERSIONS[name] = ToVersion
            _NUMPY_NDARRAY_FROM_VERSIONS[name] = FromVersion

    def RemoveNumpyNDArrayRepresentation(name: str, /) -> None:
        """"""
        global _NUMPY_NDARRAY_TO_VERSIONS, _NUMPY_NDARRAY_FROM_VERSIONS

        if name in (NUMPY, NUMPY_COMPRESSED):
            raise ValueError(
                f"{NUMPY}, {NUMPY_COMPRESSED}: Default representations cannot be removed"
            )

        del _NUMPY_NDARRAY_TO_VERSIONS[name]
        del _NUMPY_NDARRAY_FROM_VERSIONS[name]
        if name == BLOSC:
            del _NUMPY_NDARRAY_FROM_VERSIONS[BLOSC_BYTES]

    def _BloscVersion(array: array_t, /) -> tuple[str, str] | None:
        """"""
        # Do not compare packed instances of an array since blsc.pack_array(array) !=_{can be} blsc.pack_array(array)
        packed = blsc.pack_array(array)
        if isinstance(packed, bytes):
            packed = packed.decode(encoding="iso-8859-1")
            how = BLOSC_BYTES
        else:
            how = BLOSC

        return how, packed

    def _FromBloscVersion(blosc: str, /) -> array_t:
        """"""
        return blsc.unpack_array(blosc)

    def _FromBloscBytesVersion(blosc: str, /) -> array_t:
        """"""
        return blsc.unpack_array(blosc.encode(encoding="iso-8859-1"))

    def _PCABStreamVersion(array: array_t, /) -> tuple[str, str] | None:
        """"""
        if nmpy.issubclass_(
            array.dtype, (bool, nmpy.bool_, int, nmpy.integer, float, nmpy.floating)
        ):
            stream = pcst.PCA2BStream(array).decode(encoding="iso-8859-1")
            return PCA_B_STREAM, stream

        return None

    def _FromPCABStreamVersion(pca_b_stream: str, /) -> array_t:
        """"""
        return pcst.BStream2PCA(pca_b_stream.encode(encoding="iso-8859-1"))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import json
import pathlib as pthl
from typing import Any

from json_any.catalog.module import grph, nmpy, pnds, pypl, sprs
from json_any.catalog.type import (
    BUILTIN_BYTES_CONTAINERS_NAMES,
    BUILTIN_CONTAINERS_NAMES,
    JSON_TYPE,
    JSON_TYPE_NUMPY_SCALAR,
    JSON_TYPE_PREFIX_PATHLIB,
    ContainerWithName,
    TypeIsInModule,
    TypeNameOf,
    date_t,
    date_time_t,
    enum_t,
    io_bytes_t,
    named_tuple_t,
    py_array_t,
    time_delta_t,
    time_t,
    time_zone_t,
    uuid_t,
)
from json_any.constant import (
    CUSTOM_PREFIX,
    DATACLASS_PREFIX,
    NEW_FROM_JSON_DESCRIPTION,
    STANDARD_PREFIX,
    UNHANDLED_PREFIX,
)
from json_any.extension.numpy import AsNumpyArray
from json_any.extension.type import (
    TypeFromJsonType,
    builders_h,
    description_h,
    unfound_t,
)


def ObjectFromJsonString(
    jsoned: str,
    /,
    *,
    builders: builders_h = None,
    should_continue_on_error: bool = False,
) -> Any:
    """"""
    return _ObjectFromUnJSONed(
        json.loads(jsoned),
        highest_level_call=True,
        builders=builders,
        should_continue_on_error=should_continue_on_error,
    )


def _ObjectFromUnJSONed(
    description: description_h,
    /,
    *,
    highest_level_call: bool = False,
    builders: builders_h = None,
    should_continue_on_error: bool = False,
) -> Any:
    """"""
    json_type, instance = description
    if json_type.startswith(UNHANDLED_PREFIX):
        print(
            f"{json_type[UNHANDLED_PREFIX.__len__():]}: UnJSONable type. Returning None."
        )
        return None

    if builders is None:
        builders = {}

    if json_type.startswith(CUSTOM_PREFIX):
        qualified_type = json_type[CUSTOM_PREFIX.__len__() :]
        if qualified_type in builders:
            Rebuilt = builders[qualified_type]
        elif highest_level_call and ("" in builders):
            Rebuilt = builders[""]
        else:
            output_t = TypeFromJsonType(
                qualified_type, should_continue_on_error=should_continue_on_error
            )
            Rebuilt = getattr(output_t, NEW_FROM_JSON_DESCRIPTION, None)
        if Rebuilt is None:
            output = None
            print(f"{json_type}: Unfound type or type without builder. Returning None.")
        else:
            output = Rebuilt(
                _ObjectFromUnJSONed(
                    instance,
                    builders=builders,
                    should_continue_on_error=should_continue_on_error,
                )
            )
    elif json_type.startswith(DATACLASS_PREFIX):
        output_t = TypeFromJsonType(
            json_type,
            prefix=DATACLASS_PREFIX,
            should_continue_on_error=should_continue_on_error,
        )
        if output_t is unfound_t:
            output = None
            print(f"{json_type}: Unfound type. Returning None.")
        else:
            unjsoned = _ObjectFromUnJSONed(
                instance,
                builders=builders,
                should_continue_on_error=should_continue_on_error,
            )

            kwargs = {}
            for field in dtcl.fields(output_t):
                if field.init:
                    # This could be limited to init fields without default values. However, all kind of things can happen in
                    # init or post_init, so hopefully, the choice to ignore default values here works...
                    name = field.name
                    kwargs[name] = unjsoned[name]
            output = output_t(**kwargs)
            # Despite initial values being passed above, all the fields are reset here to their values at JSONing time,
            # again in case things happen in init or post_init.
            for key, value in unjsoned.items():
                setattr(output, key, value)
    elif json_type in BUILTIN_BYTES_CONTAINERS_NAMES:
        output_t = ContainerWithName(json_type)
        output = instance.encode(encoding="iso-8859-1")
        if type(output) is not output_t:
            # In practice, this is only for bytearray since bytes are already bytes.
            output = output_t(output)
    elif json_type == complex.__name__:
        output = complex(*instance)
    elif json_type == slice.__name__:
        output = slice(*instance)
    elif json_type == JSON_TYPE[uuid_t]:
        output = uuid_t(int=instance)
    elif json_type == JSON_TYPE[date_t]:
        output = date_t(*instance)
    elif json_type == JSON_TYPE[time_delta_t]:
        output = time_delta_t(*instance)
    elif json_type == JSON_TYPE[py_array_t]:
        as_list, typecode = instance
        output = py_array_t(typecode)
        output.fromlist(as_list)
    elif json_type == JSON_TYPE[io_bytes_t]:
        output = io_bytes_t(instance.encode(encoding="iso-8859-1"))
    elif json_type.startswith(JSON_TYPE_PREFIX_PATHLIB):
        type_name = json_type[JSON_TYPE_PREFIX_PATHLIB.__len__() :]
        output_t = getattr(pthl, type_name)
        output = output_t(instance)
    elif TypeIsInModule(json_type, sprs):
        type_name = TypeNameOf(json_type)
        output_t = getattr(sprs, type_name)
        output = output_t(AsNumpyArray(*instance))
    elif (json_type != JSON_TYPE_NUMPY_SCALAR) and TypeIsInModule(json_type, nmpy):
        output = AsNumpyArray(*instance)
    elif TypeIsInModule(json_type, pnds):
        type_name = TypeNameOf(json_type)
        unjsoned = _ObjectFromUnJSONed(
            instance,
            builders=builders,
            should_continue_on_error=should_continue_on_error,
        )
        output_t = getattr(pnds, type_name)
        output = output_t(unjsoned)
    elif json_type == JSON_TYPE_NUMPY_SCALAR:
        dtype, value = instance
        if isinstance(value, list):
            output = nmpy.dtype(dtype).type(complex(*value))
        else:
            output = nmpy.dtype(dtype).type(value)
    elif TypeIsInModule(json_type, pypl):
        fake_file = io_bytes_t(instance.encode(encoding="iso-8859-1"))
        image = pypl.imread(fake_file)
        fake_file.close()
        output, axes = pypl.subplots()
        axes.set_axis_off()
        axes.matshow(image)
    elif json_type == JSON_TYPE[date_time_t]:
        date, time = _ObjectFromUnJSONed(instance)
        output = date_time_t(
            date.year,
            date.month,
            date.day,
            time.hour,
            time.minute,
            time.second,
            time.microsecond,
            time.tzinfo,
            fold=time.fold,
        )
    elif json_type == JSON_TYPE[time_t]:
        time_zone = _ObjectFromUnJSONed(instance[4])
        kwargs = dict(
            zip(
                ("hour", "minute", "second", "microsecond", "tzinfo", "fold"),
                (*instance[:4], time_zone, *instance[5:]),
            )
        )
        output = time_t(**kwargs)
    elif json_type == JSON_TYPE[time_zone_t]:
        time_delta, name = _ObjectFromUnJSONed(instance)
        output = time_zone_t(time_delta, name=name)
    elif json_type.startswith(JSON_TYPE[enum_t]):
        output_t = TypeFromJsonType(
            json_type,
            prefix=JSON_TYPE[enum_t],
            should_continue_on_error=should_continue_on_error,
        )
        if output_t is unfound_t:
            output = None
            print(f"{json_type}: Unfound type. Returning None.")
        else:
            output = output_t(
                _ObjectFromUnJSONed(
                    instance,
                    builders=builders,
                    should_continue_on_error=should_continue_on_error,
                )
            )
    elif json_type.startswith(JSON_TYPE[named_tuple_t]):
        output_t = TypeFromJsonType(
            json_type,
            prefix=JSON_TYPE[named_tuple_t],
            should_continue_on_error=should_continue_on_error,
        )
        if output_t is unfound_t:
            output = None
            print(f"{json_type}: Unfound type. Returning None.")
        else:
            output = output_t._make(
                _ObjectFromUnJSONed(
                    instance,
                    builders=builders,
                    should_continue_on_error=should_continue_on_error,
                )
            )
    elif json_type in BUILTIN_CONTAINERS_NAMES:
        iterator = (
            _ObjectFromUnJSONed(
                _elm,
                builders=builders,
                should_continue_on_error=should_continue_on_error,
            )
            for _elm in instance
        )
        output_t = ContainerWithName(json_type)
        output = output_t(iterator)
    elif json_type == dict.__name__:
        output = {
            _ObjectFromUnJSONed(
                json.loads(_key),
                builders=builders,
                should_continue_on_error=should_continue_on_error,
            ): _ObjectFromUnJSONed(
                _vle,
                builders=builders,
                should_continue_on_error=should_continue_on_error,
            )
            for _key, _vle in instance.items()
        }
    elif TypeIsInModule(json_type, grph):
        type_name = TypeNameOf(json_type)
        output_t = getattr(grph, type_name)

        edges_w_attributes = _ObjectFromUnJSONed(
            instance,
            builders=builders,
            should_continue_on_error=should_continue_on_error,
        )
        attributes = {}
        edges = {}
        for node, (node_attributes, edge) in edges_w_attributes.items():
            attributes[node] = node_attributes
            edges[node] = edge

        output = grph.from_dict_of_dicts(
            edges,
            create_using=output_t,
            multigraph_input=output_t in (grph.MultiGraph, grph.MultiDiGraph),
        )
        grph.set_node_attributes(output, attributes)
    elif json_type.startswith(STANDARD_PREFIX):
        output = instance
    else:
        raise ValueError(f"{json_type}: Invalid JSON type or error while unJSONing.")

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import json
from typing import Any, Sequence

from json_any.catalog.module import grph
from json_any.catalog.type import (
    BUILTIN_BYTES_CONTAINERS,
    BUILTIN_CONTAINERS,
    JSON_TYPE,
    JSON_TYPE_NUMPY_SCALAR,
    JSON_TYPE_PREFIX_PATHLIB,
    MATPLOTLIB_CLASSES,
    NETWORKX_CLASSES,
    NUMPY_ARRAY_CLASSES,
    PANDAS_CLASSES,
    SCIPY_ARRAY_CLASSES,
    date_t,
    date_time_t,
    enum_t,
    io_bytes_t,
    named_tuple_t,
    path_t,
    py_array_t,
    time_delta_t,
    time_t,
    time_zone_t,
    uuid_t,
)
from json_any.constant import (
    CUSTOM_PREFIX,
    DATACLASS_PREFIX,
    DESCRIPTION_FOR_JSON,
    JSONING_ERROR_MARKER,
    STANDARD_PREFIX,
    UNHANDLED_PREFIX,
)
from json_any.extension.numpy import (
    AsMostConciseRepresentation,
    AsScalarRepresentation,
    IsNumpyScalar,
)
from json_any.extension.type import (
    IsFullyDataclassBased,
    IsNamedTuple,
    QualifiedType,
    TypeOfInstance,
    description_h,
    descriptors_h,
)

_HISTORY_INDENTATION = 4 * " "


def JsonStringOf(
    instance: Any, /, *, descriptors: descriptors_h = None
) -> tuple[str, Sequence[str] | None]:
    """"""
    history = []
    jsonable = _JsonableVersionOf(
        instance, history, "", highest_level_call=True, descriptors=descriptors
    )
    if all(JSONING_ERROR_MARKER not in _elm for _elm in history):
        history = None

    return json.dumps(jsonable), history


def _JsonableVersionOf(
    instance: Any,
    history: list[str],
    history_level: str,
    /,
    *,
    highest_level_call: bool = False,
    descriptors: descriptors_h = None,
) -> description_h:
    """"""
    instance_type = TypeOfInstance(instance)
    json_type = JSON_TYPE.get(instance_type, None)
    qualified_type = QualifiedType(instance_type)

    if descriptors is None:
        descriptors = {}
    if qualified_type in descriptors:
        DescriptionForJSON = descriptors[qualified_type]
    elif highest_level_call and ("" in descriptors):
        # Empty key "" is equivalent to QualifiedType(instance) when instance is
        # the object passed to "JsonStringOf". It allows to avoid to have to call
        # QualifiedType in the first place.
        DescriptionForJSON = descriptors[""]
    else:
        DescriptionForJSON = getattr(instance_type, DESCRIPTION_FOR_JSON, None)
    if DescriptionForJSON is not None:
        json_type = f"{CUSTOM_PREFIX}{qualified_type}"
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            DescriptionForJSON(instance),
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif dtcl.is_dataclass(instance):
        if IsFullyDataclassBased(instance_type):
            json_type = f"{DATACLASS_PREFIX}{qualified_type}"
            # Do not use dtcl.asdict(instance) since it recurses into dataclass
            # instances which, if they extend a "container" class like list or dict,
            # might lose information.
            description = {
                _fld.name: getattr(instance, _fld.name)
                for _fld in dtcl.fields(instance)
            }
            history.append(f"{history_level}{json_type}[{description.__len__()}]")
            jsonable = _JsonableVersionOf(
                description,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
        else:
            json_type = f"{UNHANDLED_PREFIX}{DATACLASS_PREFIX}{qualified_type}"
            jsonable = None
            history.append(
                f"{history_level}{JSONING_ERROR_MARKER}"
                f"{instance_type.__name__}: Dataclass with inheritance "
                f'without a "{DESCRIPTION_FOR_JSON}" method. Using None.'
            )
    elif instance_type in BUILTIN_BYTES_CONTAINERS:
        # bytes.hex was initially used in place of bytes.decode.
        json_type = instance_type.__name__
        jsonable = instance.decode(encoding="iso-8859-1")
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is complex:
        json_type = complex.__name__
        jsonable = (instance.real, instance.imag)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is slice:
        json_type = slice.__name__
        jsonable = (instance.start, instance.stop, instance.step)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is uuid_t:
        jsonable = instance.int
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is date_t:
        jsonable = (instance.year, instance.month, instance.day)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is time_delta_t:
        jsonable = (instance.days, instance.seconds, instance.microseconds)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is py_array_t:
        jsonable = (instance.tolist(), instance.typecode)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is io_bytes_t:
        # Buffer is assumed to be open (i.e. no instance.closed check).
        jsonable = instance.getvalue().decode(encoding="iso-8859-1")
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif issubclass(instance_type, path_t):
        json_type = f"{JSON_TYPE_PREFIX_PATHLIB}{instance_type.__name__}"
        jsonable = str(instance)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in SCIPY_ARRAY_CLASSES:
        jsonable = AsMostConciseRepresentation(instance.toarray())
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in NUMPY_ARRAY_CLASSES:
        jsonable = AsMostConciseRepresentation(instance)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in PANDAS_CLASSES:
        jsonable = _JsonableVersionOf(
            instance.to_dict(),
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif IsNumpyScalar(instance_type):
        json_type = JSON_TYPE_NUMPY_SCALAR
        description = AsScalarRepresentation(instance, instance_type)
        jsonable = (instance.dtype.char, description)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in MATPLOTLIB_CLASSES:
        fake_file = io_bytes_t()
        instance.canvas.draw()
        instance.savefig(
            fake_file,
            bbox_inches="tight",
            pad_inches=0.0,
            transparent=True,
            dpi=200.0,
            format="png",
        )
        jsonable = fake_file.getvalue().decode(encoding="iso-8859-1")
        fake_file.close()
        history.append(history_level + json_type)
    elif instance_type is date_time_t:
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            (instance.date(), instance.timetz()),
            history,
            history_level + _HISTORY_INDENTATION,
        )
    elif instance_type is time_t:
        history.append(history_level + json_type)
        jsonable = (
            instance.hour,
            instance.minute,
            instance.second,
            instance.microsecond,
            _JsonableVersionOf(
                instance.tzinfo, history, history_level + _HISTORY_INDENTATION
            ),
            instance.fold,
        )
    elif instance_type is time_zone_t:
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            (instance.utcoffset(None), instance.tzname(None)),
            history,
            history_level + _HISTORY_INDENTATION,
        )
    elif issubclass(instance_type, enum_t):
        json_type = f"{JSON_TYPE[enum_t]}{qualified_type}"
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            instance.value,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif IsNamedTuple(instance):
        json_type = f"{JSON_TYPE[named_tuple_t]}{qualified_type}"
        description = tuple(instance)
        history.append(f"{history_level}{json_type}[{description.__len__()}]")
        jsonable = _JsonableVersionOf(
            description,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif instance_type in BUILTIN_CONTAINERS:
        json_type = instance_type.__name__
        history.append(f"{history_level}{json_type}[{instance.__len__()}]")
        jsonable = [
            _JsonableVersionOf(
                _elm,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
            for _elm in instance
        ]
    elif instance_type is dict:
        # json does not accept non-str dictionary keys, hence the json.dumps.
        json_type = dict.__name__
        history.append(f"{history_level}{json_type}[{instance.__len__()}]")
        jsonable = {
            json.dumps(
                _JsonableVersionOf(
                    _key,
                    history,
                    history_level + _HISTORY_INDENTATION,
                    descriptors=descriptors,
                )
            ): _JsonableVersionOf(
                _vle,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
            for _key, _vle in instance.items()
        }
    elif instance_type in NETWORKX_CLASSES:
        edges = grph.to_dict_of_dicts(instance)
        # /!\ Node attributes are added to the edges dictionary! This must be taken into account when deJSONing. Note
        # that several attempts to avoid this have been made, including one relying on repr(node), which is based on
        # hash(node). Since the hash function gives different results across Python sessions, this could not work.
        for node, attributes in instance.nodes(data=True):
            edges[node] = (attributes, edges[node])
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            edges,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    else:
        try:
            _ = json.dumps(instance)
            json_type = f"{STANDARD_PREFIX}{instance_type.__name__}"
            jsonable = instance
            history.append(f"{history_level}{json_type}:{jsonable}")
        except TypeError:
            json_type = f"{UNHANDLED_PREFIX}{instance_type.__name__}"
            jsonable = None
            history.append(
                f"{history_level}{JSONING_ERROR_MARKER}"
                f"{json_type[UNHANDLED_PREFIX.__len__():]}: "
                f"UnJSONable type. Using None."
            )

    return json_type, jsonable
